import sys
import threading
import time
from abc import ABC, abstractmethod
from typing import TextIO
from uuid import UUID

from atpbar.progress_report import Report


class Presentation(ABC):
    '''A base class of the progress presentation.

    A subclass of this class should implement ``_present()``.
    '''

    stdout_stderr_redirection = False

    def __init__(self) -> None:

        self.out = sys.stdout
        self.err = sys.stderr

        self.lock = threading.Lock()

        self._new_task_ids = list[UUID]()
        self._active_task_ids = list[UUID]()  # in order of arrival
        self._finishing_task_ids = list[UUID]()
        self._complete_task_ids = list[UUID]()  # in order of completion
        self._report_dict = dict[UUID, Report]()

        self.interval = 1.0  # [second]
        self.last_time = time.time()

    def active(self) -> bool:
        if self._active_task_ids:
            return True
        return False

    def present(self, report: Report) -> None:
        with self.lock:
            if not self._register_report(report):
                return
            if not self._need_to_present():
                return
            self._present(report)
            self._update_registry()
            self.last_time = time.time()

    @abstractmethod
    def _present(self, report: Report) -> None:
        pass

    def _register_report(self, report: Report) -> bool:

        taskid = report['taskid']

        if taskid in self._complete_task_ids:
            return False

        self._report_dict[taskid] = report

        if taskid in self._finishing_task_ids:
            return True

        if report['last']:
            try:
                self._active_task_ids.remove(taskid)
            except ValueError:
                pass

            try:
                self._new_task_ids.remove(taskid)
            except ValueError:
                pass

            self._finishing_task_ids.append(taskid)

            return True

        if taskid in self._active_task_ids:
            return True

        if taskid in self._new_task_ids:
            return True

        self._new_task_ids.append(taskid)
        return True

    def _update_registry(self) -> None:
        self._active_task_ids.extend(self._new_task_ids)
        del self._new_task_ids[:]

        self._complete_task_ids.extend(self._finishing_task_ids)
        del self._finishing_task_ids[:]

    def _need_to_present(self) -> bool:

        if self._new_task_ids:
            return True

        if self._finishing_task_ids:
            return True

        if time.time() - self.last_time > self.interval:
            return True

        return False

    def stdout_write(self, s: str) -> None:
        with self.lock:
            self._stdout_write(s)

    def stderr_write(self, s: str) -> None:
        with self.lock:
            self._stderr_write(s)

    def _stdout_write(self, s: str) -> None:
        self._write(s, out=self.out)

    def _stderr_write(self, s: str) -> None:
        self._write(s, out=self.err)

    def _write(self, s: str, out: TextIO) -> None:
        out.write(s.rstrip())
        out.write('\n')
        out.flush()



try:
    import ipywidgets as widgets
except ImportError:
    widgets = None

try:
    from IPython.display import display
except ImportError:
    display = None  # type: ignore

try:
    from IPython import get_ipython
except ImportError:
    get_ipython = None  # type: ignore


from .spy import is_spyder_ide


def is_jupyter_notebook() -> bool:
    '''Tests if on Jupyter Notebook

    Returns
    -------
    bool
        True if on Jupyter Notebook

    '''

    if widgets is None:
        return False

    if display is None:
        return False

    try:
        if 'IPKernelApp' not in get_ipython().config:
            return False
    except:
        return False

    if is_spyder_ide():
        return False

    return True



import threading
import time
from multiprocessing import Queue
from typing import TYPE_CHECKING

from .report import Report

if TYPE_CHECKING:
    from atpbar.presentation import Presentation


class ProgressReportPickup(threading.Thread):
    '''A pickup of progress reports.

    This class picks up progress reports and presents them.

    Parameters
    ----------
    queue : multiprocessing.Queue
        The queue through which this class receives progress reports
    presentation :
        The presentation of the reports
    '''

    def __init__(self, queue: 'Queue[Report]', presentation: 'Presentation') -> None:
        super().__init__(daemon=True)
        # The daemon makes the functions registered at atexit called
        # even if the pickup is still running

        self.queue = queue
        self.presentation = presentation
        self.last_wait_time = 1.0  # [second]

        self.start()

    def end(self) -> None:
        '''end the thread'''
        self.queue.put(None)  # type: ignore
        self.join()

    def run(self) -> None:
        try:
            self._run_until_the_end_order_arrives()
            self._run_until_reports_stop_coming()
        except EOFError:
            pass

    def _run_until_the_end_order_arrives(self) -> None:
        end_order_arrived = False
        while True:
            while not self.queue.empty():
                report = self.queue.get()
                if report is None:  # the end order
                    end_order_arrived = True
                    continue
                self._process_report(report)
            if end_order_arrived:
                break
            else:
                self._short_sleep()

    def _run_until_reports_stop_coming(self) -> None:
        self.last_time = time.time()
        while self.presentation.active():
            if time.time() - self.last_time > self.last_wait_time:
                break
            while not self.queue.empty():
                report = self.queue.get()
                if report is None:
                    continue
                self.last_time = time.time()
                self._process_report(report)
            self._short_sleep()

    def _process_report(self, report: Report) -> None:
        self.presentation.present(report)

    def _short_sleep(self) -> None:
        '''sleep very briefly

        used to prevent the empty `while` loop from increasing CPU
        loads

        '''
        time.sleep(0.001)



import time
from multiprocessing import Queue
from typing import TYPE_CHECKING
from uuid import UUID

from .complement import ProgressReportComplementer
from .report import Report

if TYPE_CHECKING:
    from atpbar.stream import StreamQueue

DEFAULT_INTERVAL = 0.1  # [second]


class ProgressReporter:
    '''A progress reporter

    This class sends progress reports. The reports will be picked up
    by the pickup (`ProgressReportPickup`), which uses the reports,
    for example, to update `ProgressBar` on the screen.

    An instance of this class is initialized with a message queue::

        reporter = ProgressReporter(queue)

    The pickup, which is running in a sub-thread of the main process,
    needs to have the same queue.

    A report can be sent as::

        reporter.report(report)

    This method can be frequently called multiple times. However,
    after sending one report, the reporter wait for a certain
    ``interval`` (0.1 seconds by default) before sending another
    report of the same task. Reports from the same task received
    within this interval will be discarded. The exception for this is
    the last report. The last report, which indicates the completion
    of the task, will be always sent to the progress monitor
    regardless of whether it is given within the interval.

    Parameters
    ----------
    queue : multiprocessing.Queue
        The queue through which this class sends progress reports.
    '''

    def __init__(
        self,
        queue: 'Queue[Report]',
        notices_from_sub_processes: 'Queue[bool]',
        stream_queue: 'StreamQueue',
    ) -> None:
        self.queue = queue
        self.interval = DEFAULT_INTERVAL  # [second]
        self.last_time = dict[UUID, float]()
        self.complete_report = ProgressReportComplementer()
        self.notices_from_sub_processes = notices_from_sub_processes
        self.stream_queue = stream_queue
        self.stream_redirection_enabled = True

    def __repr__(self) -> str:
        return '{}(queue={!r}, interval={!r})'.format(
            self.__class__.__name__, self.queue, self.interval
        )

    def report(self, report: Report) -> None:
        '''send ``report`` to a progress monitor

        Parameters
        ----------
        report : ProgressReport
            a progress report

        '''

        self.complete_report(report)

        if not self._need_to_report(report):
            return

        self.queue.put(report)

        self.last_time[report['taskid']] = time.time()

    def _need_to_report(self, report: Report) -> bool:

        if report['first']:
            return True

        if report['last']:
            return True

        if report['taskid'] not in self.last_time:
            return True

        if time.time() - self.last_time[report['taskid']] > self.interval:
            return True

        return False



from typing import TypedDict
from uuid import UUID


class Report(TypedDict, total=False):
    '''Progress report

    Parameters
    ----------
    taskid : immutable
        The unique task ID.
    done : int
        The number of the iterations done so far
    total : int
            The total iterations to be done
    name : str
            A name of the task. It will be use as the label on the
            progress bars.
    first : bool
            `True` if the first report for the task. If not given,
            automatically determined from `done`; `True` if `done` is
            0, `False` otherwise
    last : bool
            `True` if the last report for the task. If not given,
            automatically determined from `done` and `total`; `True`
            if `done` equals `total`, `False` otherwise

    Notes:
    ------
    TODO: Use typing.Required for Python 3.11 and above
    '''

    taskid: UUID
    done: int
    total: int
    name: str
    first: bool
    last: bool



import shutil
from typing import TextIO
from uuid import UUID

from .base import Presentation, Report

MINIMUM_TERMINAL_WIDTH = 90


class ProgressBar(Presentation):
    stdout_stderr_redirection = True

    def __init__(self) -> None:
        super().__init__()
        self.interval = 0.1  # [second]
        self.width = self._get_width()

        self.active_bars = list[str]()
        self.just_finished_bars = list[str]()

    def __repr__(self) -> str:
        return '{}()'.format(self.__class__.__name__)

    def _get_width(self) -> int:
        try:
            columns = shutil.get_terminal_size().columns
            return max(MINIMUM_TERMINAL_WIDTH, columns - 1)
        except AttributeError:
            return MINIMUM_TERMINAL_WIDTH

    def _present(self, report: Report) -> None:
        self._erase_active_bars()
        self._compose_just_finished_bars()
        self._compose_active_bars()
        self._draw_just_finished_bars()
        self._draw_active_bars()

    def _write(self, s: str, out: TextIO) -> None:
        if not s:
            return
        self._erase_active_bars()
        out.write(s.rstrip())
        out.write('\n')
        out.flush()
        self._draw_active_bars()

    def _erase_active_bars(self) -> None:
        n_lines = len(self._active_task_ids) + len(self._finishing_task_ids)
        # must be the same as len(self.active_bars)

        if n_lines == 0:
            return

        code = '\033[1G' + '\033[A' * (n_lines - 1) + '\033[0J'
        # '\033[1G' move the cursor to the beginning of the line
        # '\033[A' move the cursor up
        # '\033[0J' clear from cursor to end of screen

        self.out.write(code)
        self.out.flush()

    def _compose_just_finished_bars(self) -> None:
        self.just_finished_bars = [
            self._compose_bar_from_taskid(i) for i in self._finishing_task_ids
        ]

    def _compose_active_bars(self) -> None:
        self.active_bars = [
            self._compose_bar_from_taskid(i)
            for i in self._active_task_ids + self._new_task_ids
        ]

    def _compose_bar_from_taskid(self, taskid: UUID) -> str:
        report = self._report_dict[taskid]
        return self._compose_bar_from_report(report)

    def _compose_bar_from_report(self, report: Report) -> str:

        percent = float(report['done']) / report['total'] if report['total'] > 0 else 1
        # e.g., 0.7143369818769065

        bar = (':' * int(percent * 40)).ljust(40, ' ')
        # e.g., '::::::::::::::::::::::::::::            '

        percent = round(percent * 100, 2)
        # e.g., 71.43

        format = ' {percent:6.2f}% {bar:s} | {done:8d} / {total:8d} |:  {name} '

        ret = format.format(
            percent=percent,
            bar=bar,
            done=report['done'],
            total=report['total'],
            name=report['name'],
        )
        # e.g., '  71.43% ::::::::::::::::::::::::::::             |     3981 /     5573 |:  task name '

        ret = ret[: self.width].ljust(self.width, ' ')
        # e.g., '  71.43% ::::::::::::::::::::::::::::             |     3981 /     5573 |:  task na'

        return ret

    def _draw_just_finished_bars(self) -> None:
        if self.just_finished_bars:
            self.out.write('\n'.join(self.just_finished_bars) + '\n')
            self.out.flush()

    def _draw_active_bars(self) -> None:
        if self.active_bars:
            self.out.write('\n'.join(self.active_bars))
            self.out.flush()



import os

try:
    import spyder  # type: ignore
except ImportError:
    spyder = None

try:
    from IPython import get_ipython
except ImportError:
    get_ipython = None  # type: ignore


def is_spyder_ide() -> bool:
    '''Tests if on Spyder IDE

    Returns
    -------
    bool
        True if on Spyder IDE

    '''

    if spyder is None:
        return False

    try:
        if 'IPKernelApp' not in get_ipython().config:
            return False
    except:
        return False

    if 'SPYDER_ARGS' not in os.environ:
        return False

    min_n_spy_var = 15  # A possible minimum number of the environmental
    # variables that start with 'SPY_' on Spyder IDE.
    n_spy_var = len([k for k in os.environ.keys() if k.startswith('SPY_')])
    if n_spy_var < min_n_spy_var:
        return False

    return True



import atexit
import contextlib
import multiprocessing
from collections.abc import Iterator

from .machine import StateMachine
from .progress_report import ProgressReporter

_machine = StateMachine()


def find_reporter() -> ProgressReporter | None:
    '''returns the progress reporter

    This function is to be called in the main process of a
    multiprocessing program. The reporter should be registered in
    sub-processes with the function `register_reporter()`

    Returns
    -------
    object
        The progress reporter

    '''
    return _machine.find_reporter()


def register_reporter(reporter: ProgressReporter) -> None:
    '''registers a reporter

    This function is to be called in sub-processes of a
    multiprocessing program.

    Parameters
    ----------
    reporter : object
        The reporter obtained in the main process by the function
        `find_reporter()`


    Returns
    -------
    None

    '''
    _machine.register_reporter(reporter)


def flush() -> None:
    '''flushes progress bars

    This function flushes all active progress bars. It returns when
    the progress bars finish updating.

    Returns
    -------
    None

    '''
    _machine.flush()


@contextlib.contextmanager
def flushing() -> Iterator[None]:
    '''Flushes progress bars on exit'''
    try:
        yield
    finally:
        flush()


def disable() -> None:
    '''disables progress bars

    This function needs to be called in the main process before
    `atpbar()` or `find_reporter()` is used.

    Returns
    -------
    None

    '''
    _machine.disable()


def shutdown() -> None:
    '''shutdowns the progress bars

    Returns
    -------
    None

    '''
    _machine.shutdown()


# This import prevents the issue
# https://github.com/alphatwirl/atpbar/issues/4
import multiprocessing.queues  # noqa: E402 F401

atexit.register(shutdown)


@contextlib.contextmanager
def fetch_reporter() -> Iterator[ProgressReporter | None]:
    yield from _machine.fetch_reporter()



from collections.abc import Iterator
from multiprocessing import Queue
from threading import Lock, current_thread, main_thread

from .presentation import create_presentation
from .progress_report import ProgressReporter, ProgressReportPickup, Report
from .stream import StreamQueue, StreamRedirection, register_stream_queue


class StateMachine:
    def __init__(self) -> None:
        self.lock = Lock()
        self.state: State = Initial()

    def find_reporter(self) -> ProgressReporter | None:
        with self.lock:
            self.state = self.state.prepare_reporter()
        return self.state.reporter

    def register_reporter(self, reporter: ProgressReporter) -> None:
        self.state = self.state.register_reporter(reporter)

    def flush(self) -> None:
        with self.lock:
            self.state = self.state.flush()

    def disable(self) -> None:
        self.state = self.state.disable()

    def shutdown(self) -> None:
        with self.lock:
            self.state = self.state.shutdown()

    def fetch_reporter(self) -> Iterator[ProgressReporter | None]:
        with self.lock:
            self.state = self.state.prepare_reporter()
        yield from self.state.fetch_reporter(lock=self.lock)


class State:
    '''The base class of the states'''

    def __init__(self) -> None:
        self.reporter: ProgressReporter | None = None

    def prepare_reporter(self) -> 'State':
        return self

    def register_reporter(self, reporter: ProgressReporter) -> 'State':
        return Registered(reporter=reporter)

    def disable(self) -> 'State':
        return Disabled()

    def fetch_reporter(self, lock: Lock) -> Iterator[ProgressReporter | None]:
        yield None

    def flush(self) -> 'State':
        return self

    def shutdown(self) -> 'State':
        return self


class Initial(State):
    '''Initial state

    The pickup is not running.
    '''

    def __init__(self) -> None:
        self.reporter = None

    def prepare_reporter(self) -> State:
        return Active()

    def fetch_reporter(self, lock: Lock) -> Iterator[ProgressReporter | None]:
        yield self.reporter

    def flush(self) -> State:
        return Active()


class Active(State):
    '''Active state

    The pickup started and is running, typically, in the main process
    '''

    def __init__(self) -> None:

        self.queue: Queue[Report] = Queue()
        self.notices_from_sub_processes: Queue[bool] = Queue()
        self.stream_queue: StreamQueue = Queue()
        self.reporter = ProgressReporter(
            queue=self.queue,
            notices_from_sub_processes=self.notices_from_sub_processes,
            stream_queue=self.stream_queue,
        )

        self.reporter_yielded = False

        self._start_pickup()

        if self.stream_redirection.disabled:
            self.reporter.stream_redirection_enabled = False

    def _start_pickup(self) -> None:
        presentation = create_presentation()
        self.pickup = ProgressReportPickup(self.queue, presentation)

        self.stream_redirection = StreamRedirection(
            queue=self.stream_queue, presentation=presentation
        )
        self.stream_redirection.start()

    def _end_pickup(self) -> None:
        self.pickup.end()

        self.stream_redirection.end()

    def _restart_pickup(self) -> None:
        self._end_pickup()
        self._start_pickup()

    def fetch_reporter(self, lock: Lock) -> Iterator[ProgressReporter | None]:

        if not in_main_thread():
            self.to_restart_pickup = False
            yield self.reporter
            return

        if self.reporter_yielded:
            # called from an inner loop
            yield self.reporter
            return

        self.reporter_yielded = True
        self.to_restart_pickup = True

        while not self.notices_from_sub_processes.empty():
            _ = self.notices_from_sub_processes.get()

        try:
            yield self.reporter
        finally:
            self.reporter_yielded = False

            while not self.notices_from_sub_processes.empty():
                _ = self.notices_from_sub_processes.get()
                self.to_restart_pickup = False

            if not self.to_restart_pickup:
                return

            with lock:
                self._restart_pickup()

    def flush(self) -> State:
        self._restart_pickup()
        return self

    def shutdown(self) -> State:
        self._end_pickup()
        return Initial()


class Registered(State):
    '''Registered state

    Typically, in a sub-process. The reporter, which has been created
    in the main process, is registered in the sub-process
    '''

    def __init__(self, reporter: ProgressReporter | None) -> None:
        self.reporter = reporter
        if reporter is None:
            return
        if reporter.stream_redirection_enabled:
            register_stream_queue(reporter.stream_queue)

    def fetch_reporter(self, lock: Lock) -> Iterator[ProgressReporter | None]:
        if self.reporter is None:
            yield self.reporter
            return

        self.reporter.notices_from_sub_processes.put(True)
        yield self.reporter


class Disabled(State):
    '''Disabled state'''

    def __init__(self) -> None:
        self.reporter = None


def in_main_thread() -> bool:
    '''test if in the main thread'''
    return current_thread() == main_thread()



__all__ = [
    'atpbar',
    'find_reporter',
    'register_reporter',
    'flush',
    'flushing',
    'disable',
    '__version__',
]
from .__about__ import __version__
from .funcs import disable, find_reporter, flush, flushing, register_reporter
from .main import atpbar



from uuid import UUID

from .report import Report


class ProgressReportComplementer:
    '''Complement progress reports

    Complement a progress report with the previous report for the same
    task.

    Parameters
    ----------
    report : Report
        A report must always include `taskid`. The first report for a task must
        include `done`, `total`, and 'name'. The `first` and `last` will be
        automatically determined if not given.

    '''

    def __init__(self) -> None:
        self.previous_reports = dict[UUID, Report]()

    def __call__(self, report: Report) -> None:
        taskid = report['taskid']
        if taskid in self.previous_reports:
            self._complement(taskid, report)
        self._first(report)
        self._last(report)
        self._store(taskid, report.copy())

    def _complement(self, taskid: UUID, report: Report) -> None:
        report_copy = report.copy()
        report.update(Report(taskid=taskid))
        report.update(self.previous_reports[taskid])
        report.update(report_copy)

    def _first(self, report: Report) -> None:
        if 'first' in report:
            return
        report['first'] = report['done'] == 0

    def _last(self, report: Report) -> None:
        if 'last' in report:
            return
        report['last'] = report['done'] >= report['total']

    def _store(self, taskid: UUID, report: Report) -> None:
        report.pop('first', None)
        report.pop('last', None)
        self.previous_reports[taskid] = report



import sys
import threading
from collections.abc import Callable
from enum import Enum
from io import TextIOBase
from multiprocessing import Queue
from typing import Any, TypeAlias

from .presentation import Presentation


class FD(Enum):
    STDOUT = 1
    STDERR = 2


StreamQueue: TypeAlias = 'Queue[tuple[str, FD] | None]'


class StreamRedirection:
    def __init__(self, queue: StreamQueue, presentation: Presentation) -> None:
        self.disabled = not presentation.stdout_stderr_redirection
        if self.disabled:
            return

        self.queue = queue
        self.presentation = presentation

        self.stdout = Stream(self.queue, FD.STDOUT)
        self.stderr = Stream(self.queue, FD.STDERR)

    def start(self) -> None:
        if self.disabled:
            return

        self.pickup = StreamPickup(
            self.queue, self.presentation.stdout_write, self.presentation.stderr_write
        )
        self.pickup.start()

        self.stdout_org = sys.stdout
        sys.stdout = self.stdout  # type: ignore

        self.stderr_org = sys.stderr
        sys.stderr = self.stderr  # type: ignore

    def end(self) -> None:
        if self.disabled:
            return

        sys.stdout = self.stdout_org
        sys.stderr = self.stderr_org
        self.queue.put(None)
        self.pickup.join()


def register_stream_queue(queue: StreamQueue) -> None:
    if queue is None:
        return
    sys.stdout = Stream(queue, FD.STDOUT)  # type: ignore
    sys.stderr = Stream(queue, FD.STDERR)  # type: ignore


class Stream(TextIOBase):
    def __init__(self, queue: StreamQueue, fd: FD) -> None:
        self.fd = fd
        self.queue = queue
        self.buffer = ''

    def write(self, s: str) -> int:
        # sys.__stdout__.write(repr(s))
        # sys.__stdout__.write('\n')

        try:
            endswith_n = s.endswith('\n')
        except:
            self.flush()
            self.queue.put((s, self.fd))
            return len(s)

        if endswith_n:
            self.buffer += s
            self.flush()
            return len(s)

        self.buffer += s
        return len(s)

    def flush(self) -> None:
        if not self.buffer:
            return
        self.queue.put((self.buffer, self.fd))
        self.buffer = ''


class StreamPickup(threading.Thread):
    def __init__(
        self,
        queue: StreamQueue,
        stdout_write: Callable[[str], Any],
        stderr_write: Callable[[str], Any],
    ) -> None:
        super().__init__(daemon=True)
        self.queue = queue
        self.stdout_write = stdout_write
        self.stderr_write = stderr_write

    def run(self) -> None:
        try:
            while True:
                m = self.queue.get()
                if m is None:
                    break
                s, f = m
                if f == FD.STDOUT:
                    self.stdout_write(s)
                elif f == FD.STDERR:
                    self.stderr_write(s)
                else:
                    raise ValueError('unknown fd: {!r}'.format(f))

        except EOFError:
            pass



import time
from uuid import UUID

from .base import Presentation, Report


class ProgressPrint(Presentation):
    def __init__(self) -> None:
        super().__init__()
        self.interval = 60.0  # [second]
        self.last_time_map = dict[UUID, float]()

    def __repr__(self) -> str:
        return '{}()'.format(self.__class__.__name__)

    def present(self, report: Report) -> None:

        if not self._register_report(report):
            return

        if not self._need_to_present_(report):
            return

        self._present(report)

        self.last_time_map[report['taskid']] = self._time()

    def _present(self, report: Report) -> None:
        time_ = time.strftime('%m/%d %H:%M', time.localtime(time.time()))
        percent = float(report['done']) / report['total'] if report['total'] > 0 else 1
        percent = round(percent * 100, 2)
        line = '{time} : {done:8d} / {total:8d} ({percent:6.2f}%): {name} '.format(
            time=time_,
            done=report['done'],
            total=report['total'],
            percent=percent,
            name=report['name'],
        )
        line = '{}\n'.format(line)
        self.out.write(line)
        self.out.flush()

    def _need_to_present_(self, report: Report) -> bool:

        if report['first']:
            return True

        if report['last']:
            return True

        if report['taskid'] not in self.last_time_map:
            return True

        if self._time() - self.last_time_map[report['taskid']] > self.interval:
            return True

        return False

    def _time(self) -> float:
        return time.time()



import sys

from .bartty import ProgressBar
from .base import Presentation
from .detect.jupy import is_jupyter_notebook
from .txtprint import ProgressPrint

try:
    from .barjupyter import ProgressBarJupyter
except ImportError:
    pass


def create_presentation() -> Presentation:
    '''Create a presentation of progress report, e.g., progress bars

    Returns
    -------
    object
        an instance of ProgressBar if on TTY
        an instance of ProgressBarJupyter if on Jupyter Notebook
        an instance of ProgressPrint otherwise

    '''

    if sys.stdout.isatty():
        return ProgressBar()

    if is_jupyter_notebook():
        try:
            return ProgressBarJupyter()
        except:
            pass

    return ProgressPrint()



from uuid import UUID

import ipywidgets as widgets
from IPython.display import display

from .base import Presentation, Report


class ProgressBarJupyter(Presentation):
    def __init__(self) -> None:
        super().__init__()
        self.interval = 0.05  # [second]

        self.container_widget = None
        self.active_box_list = list[widgets.HBox]()
        self.complete_box_list = list[widgets.HBox]()
        self.widget_dict = dict[
            UUID, tuple[widgets.HBox, widgets.IntProgress, widgets.HTML]
        ]()

    def __repr__(self) -> str:
        return '{}()'.format(self.__class__.__name__)

    def _present(self, report: Report) -> None:
        self._create_widgets()
        self._update_widgets()

    def _create_widgets(self) -> None:
        if self.container_widget is None:
            self.container_widget = widgets.VBox()
            display(self.container_widget)

        for taskid in self._new_task_ids:
            report = self._report_dict[taskid]
            self._create_widget(report)

    def _create_widget(self, report: Report) -> None:
        bar = widgets.IntProgress(
            value=report['done'],
            min=0,
            max=report['total'],
            description='',
            bar_style='',  # 'success', 'info', 'warning', 'danger' or ''
            orientation='horizontal',
        )
        label = widgets.HTML(value='')
        box = widgets.HBox([bar, label])
        self.active_box_list.append(box)
        if self.container_widget is not None:
            self.container_widget.children = (
                self.complete_box_list + self.active_box_list
            )
        self.widget_dict[report['taskid']] = (box, bar, label)

    def _update_widgets(self) -> None:
        for taskid in (
            self._finishing_task_ids + self._active_task_ids + self._new_task_ids
        ):
            report = self._report_dict[taskid]
            self._update_widget(report)

        self._reorder_widgets(report)

        if not self._new_task_ids and not self._active_task_ids:
            self.container_widget = None
            self.active_box_list[:] = []
            self.complete_box_list[:] = []
            self.widget_dict.clear()

    def _update_widget(self, report: Report) -> None:

        percent = float(report['done']) / report['total'] if report['total'] > 0 else 1
        percent = round(percent * 100, 2)
        percent_fmt = '{:6.2f}%'.format(percent)

        box = self.widget_dict[report['taskid']][0]
        box  # to silence not-used warning

        bar = self.widget_dict[report['taskid']][1]
        bar.value = report['done']
        bar.max = report['total']
        bar.description = percent_fmt
        if report['last']:
            bar.bar_style = 'success'

        label = self.widget_dict[report['taskid']][2]
        name_field_length = 32
        percent = float(report['done']) / report['total'] if report['total'] > 0 else 1
        bar = (':' * int(percent * 40)).ljust(40, ' ')
        percent = round(percent * 100, 2)
        name = report['name'][0:name_field_length]
        label.value = '<pre> | {:8d} / {:8d} |: {:<{}s}</pre>'.format(
            report['done'], report['total'], name, name_field_length
        )

    def _reorder_widgets(self, report: Report) -> None:
        for taskid in self._finishing_task_ids:
            box, bar, label = self.widget_dict[taskid]
            if box in self.active_box_list:
                self.active_box_list.remove(box)
                self.complete_box_list.append(box)
                if self.container_widget is not None:
                    self.container_widget.children = (
                        self.complete_box_list + self.active_box_list
                    )
